/* ast.c -- Abstract Syntax Tree for parser */

/* Copyright (C) 2003 Mike Thielman 
 *
 * This file is part of pr0nsh, the Porn Shell.
 *
 * Pr0nsh is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2, or (at your option) any later
 * version.
 *
 * Pr0nsh is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Pr0nsh; see the file COPYING.  If not, write to the Free Software
 * Foundation, 59 Temple Place, Suite 330, Boston, MA 02111 USA. */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <assert.h>

#include "parse/ast.h"
#include "util/llist.h"
#include "util/debug.h"
#include "util/symbol.h"
#include "shell/shell.h"
#include "shell/redir.h"
#include "shell/process.h"


int ASTExecute(ASTCommand *ast)
{
  REDIR *tmp;
  Symbol *s;
  int ret = RET_SUCCESS;
  char errorcode[5];
  int pipefd[2],pipefdclose[2];
  int in_pipe = 0,pipe_close = 0;
  int i;

  while(ast) {
    if(ast->redir) {
      tmp = ast->redir;
      while(tmp) {
        tmp->fdclose0 = 0;
        tmp->fdclose1 = 0;
        switch(tmp->type) {
          case REDIR_OUTPUT:
            tmp->fd = open(tmp->filename,O_WRONLY | O_CREAT | O_TRUNC,
                           S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
            if(tmp->fd < 0) {
              warning("Unable to open file for write",tmp->filename);
            }
            else {
              switch(tmp->stream) {
                case STREAM_STDIN:
                  warning("Invalid redirection","0>");
                  break;
                case STREAM_STDOUT:
                  setSTDOUTfd(tmp->fd);
                  break;
                case STREAM_STDERR:
                  setSTDERRfd(tmp->fd);
                  break;
                default:
                  break;
              }
            }
            break;
          case REDIR_OUTPUT_APPEND:
            tmp->fd = open(tmp->filename,O_WRONLY | O_CREAT | O_APPEND);
            if(tmp->fd < 0) {
              warning("Unable to open file for append",tmp->filename);
            }
            else {
              switch(tmp->stream) {
                case STREAM_STDIN:
                  warning("Invalid redirection","0>>");
                  break;
                case STREAM_STDOUT:
                  setSTDOUTfd(tmp->fd);
                  break;
                case STREAM_STDERR:
                  setSTDERRfd(tmp->fd);
                  break;
                default:
                  break;
              }
            }
            break;
          case REDIR_INPUT:
            tmp->fd = open(tmp->filename,O_RDONLY);
            if(tmp->fd < 0) {
              warning("Unable to open file for read",tmp->filename);
            }
            else {
              switch(tmp->stream) {
                case STREAM_STDIN:
                  setSTDINfd(tmp->fd);
                  break;
                case STREAM_STDOUT:
                  fatal("ASTExecute got redirection of 1<\n");
                  break;
                case STREAM_STDERR:
                  fatal("ASTExecute got redirection of 2<\n");
                  break;
                default:
                  break;
              }
            }
            break;
          default:
            fatal("Invalid redirection type %d in ASTExecute\n", tmp->type);
            break;
        } 
        doutput("REDIR stream %d type %d file %s\n",
                  tmp->stream, tmp->type, tmp->filename);
        tmp=tmp->next;
      }
    }
    if(in_pipe) {
      // Add a redir of pipefd[0] to stdin 
      if(pipefd[0] != STREAM_STDIN) {
        if(ast->redir) {
          tmp->next = newREDIR();
          tmp = tmp->next;
        } else {
          ast->redir = newREDIR();
          tmp = ast->redir;
        }
        tmp->filename = NULL;
        tmp->next = NULL;
        tmp->stream = STREAM_STDIN;
        tmp->type = REDIR_INPUT;
        tmp->fd = pipefd[0];
        tmp->fdclose0 = pipefd[1];
        tmp->fdclose1 = pipefd[0];
      }
      in_pipe = 0;
      pipe_close = 1;
      for(i = 0; i < 2; i++) pipefdclose[i] = pipefd[i];
    }
    if(ast->next) { // This is the first part of a pipe
      if (pipe(pipefd) < 0) {
        fatal("Unable to create pipe.\n");
      }
      doutput("Created pipe %d to %d cmd %s\n", pipefd[1], pipefd[0],
              ast->cmd);
      in_pipe = 1;
      // Add a redir of stdout to pipefd[1]
      if(pipefd[1] != STREAM_STDOUT) {
        if(ast->redir) {
          tmp->next = newREDIR();
          tmp = tmp->next;
        } else {
          ast->redir = newREDIR();
          tmp = ast->redir;
        }
        tmp->filename = NULL;
        tmp->next = NULL;
        tmp->stream = STREAM_STDOUT;
        tmp->type = REDIR_OUTPUT;
        tmp->fd = pipefd[1];
        tmp->fdclose0 = 0;
        tmp->fdclose1 = 0;
//        tmp->fdclose = pipefd[0];
      }
    }
    switch(ast->type) {
      case AST_FUNC:
        if(strcmp(ast->cmd,"")) {
          doutput("AST: Execute command %s (%s)\n", ast->cmd, ll_to_charp(ast->args));
          ret = executeCmd(ast);
          freell(ast->args);
        }
        break;
      case AST_HIST:
        doutput("AST: Execute history %s\n", ast->expr);
        execFromHistory(ast->expr);
        break;
      default:
        fatal("AST got invalid type %d\n", ast->type);
        break;
    }
    setSTDINfd(STREAM_STDIN);
    setSTDOUTfd(STREAM_STDOUT);
    setSTDERRfd(STREAM_STDERR);
    ast = ast->next;
  }
  // Print out any jobs whose status has changed
  _builtin_jobs(createll("-n"));
  sprintf(errorcode, "%d", ret);
  AssignVar("?",errorcode);
  return ret;
}

int getASTQuiet(ASTCommand *ast)
{
  assert(ast);
  return ast->quiet;
}

void freeAST(ASTCommand *ast)
{
  if(ast) {
    freeAST(ast->next);
    freeREDIR(ast->redir);
    freell(ast->args);
    free(ast->cmd);
    free(ast->expr);
    free(ast->var);
    free(ast);
  }
}

ASTCommand *copyAST(ASTCommand *ast)
{
  ASTCommand *copy = NULL;

  copy = newAST();
  copy->type = ast->type;
  copy->bg = ast->bg;
  copy->quiet = ast->quiet;
  if(ast->var)    copy->var = strdup(ast->var);
  else            copy->var = NULL;
  if(ast->expr)   copy->expr = strdup(ast->expr);
  else            copy->expr = NULL;
  if(ast->cmd)    copy->cmd = strdup(ast->cmd);
  else            copy->cmd = NULL;
  if(ast->args)   copy->args = copyLinkedList(ast->args);
  else            copy->args = NULL;
  if(ast->redir)  copy->redir = copyREDIR(ast->redir);
  else            copy->redir = NULL;
  if(ast->next)   copy->next = copyAST(ast->next);
  else            copy->next = NULL;

  return copy; 
}

ASTCommand *newAST(void)
{
  ASTCommand *ast;

  ast = emalloc(sizeof(ASTCommand));

  ast->var = NULL;
  ast->expr = NULL;
  ast->cmd = NULL;
  ast->args = NULL;
  ast->redir = NULL;
  ast->next = NULL;

  return ast;
}
